/**
* Copyright 2007 - 2010 Skyway Software, Inc.
*/
package org.skyway.spring.util.dao.query;

import java.beans.PropertyDescriptor;
import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.skyway.common.util.ConversionUtils;
import org.springframework.beans.BeanWrapper;
import org.springframework.beans.BeanWrapperImpl;
import org.springframework.beans.PropertyAccessorFactory;
import org.springframework.dao.DataAccessException;
import org.springframework.jdbc.core.ResultSetExtractor;

public class SQLResultsProvider implements ResultSetExtractor, Serializable {
	private static transient Log _logger = LogFactory.getLog(SQLResultsProvider.class);
	private static final long serialVersionUID = 1L;
	private static int NO_LIMIT = -1;

	private List<List<Object>> extractedResults;
	private List<String> columnNames;
	private int columnCount = -1;
	private int startResult;
	@SuppressWarnings("unused")
	private int maxResults;

	public SQLResultsProvider() {
		this(NO_LIMIT, NO_LIMIT);
	}

	public SQLResultsProvider(int startResult, int maxResults) {
		this.startResult = startResult;
		this.maxResults = maxResults;
	}

	public Object extractData(ResultSet resultSet) throws SQLException, DataAccessException {
		positionResultSet(resultSet);

		columnNames = getColumnNames(resultSet);

		extractedResults = extractRows(resultSet);

		return this;
	}

	public <T> T getSingleResult(String columnName, Class<T> targetClass) {
		List<T> results = new ArrayList<T>();
		T result = null;

		getResults(columnName, targetClass, results, 1);
		if (results.size() > 0) {
			result = results.get(0);
		}

		return result;
	}

	public <T> Set<T> getResultsAsSet(String columnName, Class<T> targetClass) {
		Set<T> results = new LinkedHashSet<T>();

		getResults(columnName, targetClass, results, NO_LIMIT);

		return results;
	}

	public <T> List<T> getResultsAsList(String columnName, Class<T> targetClass) {
		List<T> results = new ArrayList<T>();

		getResults(columnName, targetClass, results, NO_LIMIT);

		return results;
	}

	@SuppressWarnings({"nls" , "unchecked"})
	public <T> void getResults(String columnName, Class<T> targetClass, Collection<T> results, int limit) {
		List<PropertyDescriptor> propertyDescriptors;
		T targetInstance = null;
		int columnNameIndex;

		try {
			if (hasDirectDatabaseRepresentation(targetClass)) {
				for (List<Object> columnResults : extractedResults) {
					columnNameIndex = getColumnNameIndex(columnName);
					if (columnNameIndex == -1) {
						columnNameIndex = 0;
					}

					if (_logger.isDebugEnabled()) {
						_logger.debug("Using column index: " + columnNameIndex + " for column name: " + columnName);
					}

					Object columnValue = columnResults.get(columnNameIndex);
					if (_logger.isDebugEnabled()) {
						_logger.debug("Column value from database: " + columnValue);
					}

					T convertedValue = convertValue(columnValue, targetClass);

					results.add(convertedValue);
				}
			} else {
				targetInstance = targetClass.newInstance();
				propertyDescriptors = getPropertyDescriptors(targetInstance);

				for (List<Object> columnResults : extractedResults) {
					targetInstance = targetClass.newInstance();
					BeanWrapper beanWrapper = new BeanWrapperImpl(targetInstance);
					int currentColumn = 0;
					for (PropertyDescriptor propertyDescriptor : propertyDescriptors) {
						if (propertyDescriptor != null) {
							Object columnValue = columnResults.get(currentColumn);
							if (_logger.isDebugEnabled()) {
								_logger.debug("Column value from database: " + columnValue);
							}

							T convertedValue = (T)convertValue(columnValue, propertyDescriptor.getReadMethod().getReturnType()); 

							beanWrapper.setPropertyValue(propertyDescriptor.getName(), convertedValue);
						}
						currentColumn++;
					}
					results.add(targetInstance);
				}
			}
		} catch (IllegalAccessException x) {
			throw new RuntimeException("IllegalAccessException getting result value for variable: " + columnName + " of type: " + targetClass.getName(), x);
		} catch (InstantiationException x) {
			throw new RuntimeException("InstantiationException getting result value for variable: " + columnName + " of type: " + targetClass.getName(), x);
		} catch (SQLException x) {
			throw new RuntimeException("SQLException getting result value for variable: " + columnName + " of type: " + targetClass.getName(), x);
		}
	}

	protected <T> T convertValue(Object sourceValue, Class<T> targetType){
		T convertedValue = (T) ConversionUtils.convert(sourceValue, targetType);
		
		if (_logger.isDebugEnabled()) {
			_logger.debug("Converted value: " + convertedValue); //$NON-NLS-1$
		}
		
		return convertedValue;
	}
	
	protected boolean hasDirectDatabaseRepresentation(Class<?> classToCheck) {
		return classToCheck.isPrimitive() || ConversionUtils.isPrimitiveWrapper(classToCheck) || String.class.isAssignableFrom(classToCheck) || Date.class.isAssignableFrom(classToCheck) || Calendar.class.isAssignableFrom(classToCheck) || BigDecimal.class.isAssignableFrom(classToCheck) || BigInteger.class.isAssignableFrom(classToCheck);
	}

	protected List<List<Object>> extractRows(ResultSet resultSet) throws SQLException {
		List<List<Object>> rows = new ArrayList<List<Object>>();

		while (resultSet.next()) {
			rows.add(extractRow(resultSet));
		}

		return rows;
	}

	protected List<Object> extractRow(ResultSet resultSet) throws SQLException {
		List<Object> row = new ArrayList<Object>();
		Object columnResult;
		int columns;

		columns = getColumnCount(resultSet);
		for (int index = 1; index <= columns; index++) {
			columnResult = resultSet.getObject(index);
			row.add(columnResult);
		}

		return row;
	}

	protected int getColumnCount(ResultSet resultSet) throws SQLException {
		if (columnCount == -1) {
			columnCount = resultSet.getMetaData().getColumnCount();
		}

		return columnCount;
	}

	@SuppressWarnings({"nls"})
	protected List<String> getColumnNames(ResultSet resultSet) throws SQLException {
		String columnName;
		int columns;

		if (columnNames == null) {
			columns = getColumnCount(resultSet);
			columnNames = new ArrayList<String>();
			for (int index = 1; index <= columns; index++) {
				columnName = resultSet.getMetaData().getColumnLabel(index);
				columnNames.add(columnName);
			}

			if (_logger.isDebugEnabled()) {
				_logger.debug("Column Names: " + columnNames);
			}
		}

		return columnNames;
	}

	protected int getColumnNameIndex(String columnName) throws SQLException {
		int columnNameIndex = 0;

		for (String metaDataColumnName : columnNames) {
			if (metaDataColumnName.equalsIgnoreCase(columnName)) {
				return columnNameIndex;
			} else {
				columnNameIndex++;
			}
		}

		return -1;
	}

	@SuppressWarnings({"nls"})
	protected List<PropertyDescriptor> getPropertyDescriptors(Object resultObject) throws SQLException {
		BeanWrapper beanWrapper = PropertyAccessorFactory.forBeanPropertyAccess(resultObject);
		List<PropertyDescriptor> columnPropertyDescriptors;
		PropertyDescriptor[] beanPropertyDescriptors;
		PropertyDescriptor propertyDescriptor;

		if (_logger.isDebugEnabled()) {
			_logger.debug("Matching properties for class: " + resultObject.getClass().getName());
		}

		beanPropertyDescriptors = beanWrapper.getPropertyDescriptors();
		columnPropertyDescriptors = new ArrayList<PropertyDescriptor>();

		for (String columnName : columnNames) {
			propertyDescriptor = getPropertyDescriptor(beanPropertyDescriptors, columnName, false);
			if (propertyDescriptor != null) {
				if (_logger.isDebugEnabled()) {
					_logger.debug("Matching property found: " + propertyDescriptor.getName() + " for column: " + columnName);
				}
				columnPropertyDescriptors.add(propertyDescriptor);
			} else {
				columnPropertyDescriptors.add(null);
			}
		}

		return columnPropertyDescriptors;
	}

	protected PropertyDescriptor getPropertyDescriptor(PropertyDescriptor[] propertyDescriptors, String columnName, boolean caseSensitive) {
		String propertyName;

		for (PropertyDescriptor propertyDescriptor : propertyDescriptors) {
			propertyName = propertyDescriptor.getName();

			if (propertyName != null && (caseSensitive ? propertyName.equals(columnName) : propertyName.equalsIgnoreCase(columnName))) {
				return propertyDescriptor;
			}
		}

		return null;
	}

	protected void positionResultSet(ResultSet resultSet) throws SQLException {
		if (startResult > 0) {
			if (resultSet.isAfterLast() == false) {
				// The start result is 0 based while the result set is 1 based.
				// The move the current row to the value of start result because
				// that will put the current row just before the row we want to
				// start with. For example, if I wanted to get results starting
				// at row 5 of the result set, I would be passing in a 4 as the
				// startResult (because it is 0 based, 4 indicates the 5th row)
				// This would position my result set at row 4 so that when the
				// result set is processed, it is immediately moved to the next
				// row - row 5 - the row we wanted to start with.
				//
				// If startResult is 0, we don't need to do anything because the
				// resultSet is positioned before the first row.
				resultSet.absolute(startResult);
			}
		}
	}
}
