/**
* Copyright 2007 - 2010 Skyway Software, Inc.
*/
package org.skyway.spring.util.webservice;

/**
 * Holds credentials for an outgoing web service call.  The credentials are stored
 * in a ThreadLocal class and picked up by the WSSecurityOutInterceptor, the ApplyCredentialsHandler
 * and the HttpBasicAuthenticationSupplier classes.  Typical usage would look like this:
 * 
 * try{
 *      WebServiceInvocationCredentials credentials = new WebServiceInvocationCredentials(your_authentication_type, your_user_name, your_password, hash_password_flag);
 *      WebServiceInvocationCredentialsHolder.set(credentials);
 *      
 *      // Call your web service
 *      
 * }finally{
 * 		WebServiceInvocationCredentialsHolder.set(null);
 * }
 * 
 * @author jperkins
 *
 */
public class WebServiceInvocationCredentialsHolder {
	private static ThreadLocal<WebServiceInvocationCredentials> invocationThreadLocal = new ThreadLocal<WebServiceInvocationCredentials>(){
		private WebServiceInvocationCredentials credentials;
		
		@Override
		public WebServiceInvocationCredentials get(){
			return credentials;
		}
		
		@Override
		public void set(WebServiceInvocationCredentials credentials){
			this.credentials = credentials;
		}
	};
	
	/**
	 * Get the web service invocation credentials for the current thread.  Null will be returned
	 * if no credentials have been set on the current thread.  This is currently used by cxf code
	 * to authenticate outgoing web service calls.
	 * 
	 * @return
	 */
	public static WebServiceInvocationCredentials get(){
		return invocationThreadLocal.get();
	}

	/**
	 * Set the web service invocation credentials onto the current thread.  It is important to wrap
	 * this call in a try - finally statement and to set the credentials to null in the finally
	 * block.  Otherwise, the credentials will stick around on the current thread which may cause
	 * downstream effects.
	 * 
	 * @param currentParameters
	 */
	public static void set(WebServiceInvocationCredentials currentParameters){
		invocationThreadLocal.set(currentParameters);
	}
}
